//
//  NzAnalyticsAPI.m
//  NzAnalyticsDemo
//
//  Created by FY on 2022/5/13.
//  Copyright © 2022 FY. All rights reserved.
//

#import <Foundation/Foundation.h>
#import "NzAnalyticsRealAPI.h"
#import "NzIntegration.h"
#import "NzUtils.h"
#import "NzAnalyticsNetwork.h"
#import "NzFileStorage.h"
#import "NzUtm.h"
#import "const.h"

#import "sys/utsname.h"

NSString * const kFileStorageApiUtm = @"io.naza.analytics.storage.api.utm";

@interface NzAnalyticsRealAPI ()

@property(nonatomic, strong) NzAnalyticsOption * option;
@property(nonatomic, strong) NzIntegration * integration;
@property(nonatomic, strong) NzUtm * utm;

@end

@implementation NzAnalyticsRealAPI

- (instancetype)initWithOption: (NzAnalyticsOption *)option
{
    if (self = [self init]) {
        self.option = option;
        self.integration = [[NzIntegration alloc] initWithConfiguration:self.option];
        [self loadStorage];

    }
    return self;
}

#pragma mark - identity
- (void)setIdentities:(NSArray *)identities {
    [self.integration setIdentities:identities];
}

- (void)addIdentities:(NSArray *)identities {
    [self.integration addIdentities:identities];
}

- (void)cleanIdentities {
    [self.integration cleanIdentities];
}

#pragma mark - profile
- (void)setProfile:(NSDictionary *)profile {
    [self sendProfileMessageWithData:profile action:kMessageDomainContactActionSet];
}

#pragma mark - event track
- (void)track:(NzEvent *)event {
    if (event == nil) {
        return;
    }
    
    NSMutableDictionary * data = [event toDictionary];
    [self fulfillCommonEventProps:data];
    NSMutableDictionary * message = [NzUtils buildMessageWithDomain:kMessageDomainEvent action:kMessageDomainEventActionTrack data:data];
    [self.integration sendMessages:@[message]];
}

#pragma mark - other
- (void)flush {
    [self.integration flush];
}

- (NzAnalyticsOption *)option {
    return _option;
}

- (void)loadStorage
{
    self.utm = [[NzUtm alloc] initWithDictionary:[[NzFileStorage sharedInstance] dictionaryForKey:kFileStorageApiUtm]];
}


- (void)persistUtm
{
    if (self.utm != nil) {
        [[NzFileStorage sharedInstance] setDictionary:[self.utm toDictionary] forKey:kFileStorageApiUtm];
    }
}

- (void)sendProfileMessageWithData:(id)data action:(NSString *)action
{
    if (data == nil || action == nil) {
        return;
    }
    NSMutableDictionary * message = [NzUtils buildMessageWithDomain:kMessageDomainContact action:action data:data];
    [self.integration sendMessages:@[message]];
}

- (void)fulfillCommonEventProps: (NSMutableDictionary *)event
{
    UIDevice *currentDevice = [UIDevice currentDevice];
    NSDictionary * commonProps = @{
        @"sdkVersion": kSdkVersion,
        @"sdkType": kSdkType,
        @"platform": @"iOS",
        @"screenWidth": [NSNumber numberWithDouble:[[UIScreen mainScreen] bounds].size.width],
        @"screenHeight":[NSNumber numberWithDouble:[[UIScreen mainScreen] bounds].size.height],
        @"appVersion": ObjectOrNull([[NSBundle mainBundle] objectForInfoDictionaryKey: @"CFBundleShortVersionString"]),
        @"os": @"iOS",
        @"osVersion": ObjectOrNull([currentDevice systemVersion]),
        @"networkType": ObjectOrNull([self getNetworkStatus]),
        @"manufacturer": @"APPLE",
        @"deviceModel": ObjectOrNull([self deviceModel]),
        @"debugMode": [NSNumber numberWithBool:self.option.debugMode],
        @"bundleKey": ObjectOrNull([[NSBundle mainBundle] bundleIdentifier]),
        @"utm": self.utm ? ObjectOrNull([self.utm toDictionary]):ObjectOrNull(nil)
    };
    [event addEntriesFromDictionary: commonProps];
    NSNumber * eventDate = [event objectForKey:@"eventDate"];
    if (eventDate == nil) {
        eventDate = [NzUtils now];
        [event setObject:eventDate forKey:@"eventDate"];
    }
    if ([event objectForKey:@"items"] != nil) {
        NSArray * subevents = [event objectForKey:@"items"];
        for (int i = 0; i < [subevents count]; i++) {
            NSMutableDictionary * subevent = [subevents objectAtIndex:i];
            [subevent addEntriesFromDictionary: commonProps];
            [subevent setObject:eventDate forKey:@"eventDate"]; // 子事件必须和父事件的eventDate保持一致
        }
    }
}

- (NSString *)deviceModel
{
    struct utsname systemInfo;
    uname(&systemInfo);
    NSString * deviceName = [NSString stringWithCString:systemInfo.machine encoding:NSUTF8StringEncoding];
    if (!deviceName) {
        return [[UIDevice currentDevice] model];
    } else {
        return deviceName;
    }
}

- (NSString *)getNetworkStatus
{
    NzNetworkStatus status = [[NzAnalyticsNetwork shared] getNetworkStatus];
    if (status == NzReachableViaWiFi) {
        return @"WIFI";
    } else if (status == NzReachableViaWWAN) {
        return @"WWAN";
    } else {
        return @"OFFLINE";
    }
}


@end
