//
//  NzAnalyticsIntegration.m
//  NzAnalytics
//
//  Created by FY on 2018/10/12.
//  Copyright © 2018 FY. All rights reserved.
//

#import <UIKit/UIKit.h>
#import "NzIntegration.h"
#import "NzFileStorage.h"
#import "NzHttpClient.h"
#import "NzUtils.h"
#import "NzReachability.h"
#import "NzIdentity.h"
#import "NzAnalyticsNetwork.h"

#import "const.h"

NSString * const kFileStorageIdentities = @"io.naza.analytics.storage.v2.identities";   // v2结构不兼容
NSString * const kFileStorageIntegrationContext = @"io.naza.analytics.storage.v2.integration.context"; // v2结构不兼容
NSString * const kFileStorageMessages = @"io.naza.analytics.storage.v2.messages";   // v2结构不兼容

NSString * const kPathMessages = @"api/sdk/v2/batch_messages";
NSString * const kIdentityTypeAnonymousId = @"anonymousId";

@interface NzIntegration ()

@property(nonatomic, strong) NzFileStorage * storage;
@property(nonatomic, strong) NSURLSessionTask * requestTask;
@property(nonatomic, strong) NSTimer * flushTimer;
@property(nonatomic, strong) dispatch_queue_t serialQueue;
@property(nonatomic, strong) NzAnalyticsOption * configuration;
@property(nonatomic, strong) NzHttpClient * httpClient;
@property(nonatomic, strong) NSString * clientId;
@property(nonatomic, strong) NSMutableArray * cacheIdentities;
@property(nonatomic, strong) NSMutableArray * messagesQueue;
@property(nonatomic, assign) UIBackgroundTaskIdentifier backgroundFlushTask;

@end

@implementation NzIntegration

- (instancetype)initWithConfiguration:(NzAnalyticsOption *) configuration
{
    if (self = [self init]) {
        self.configuration = configuration;
        
        self.storage = [[NzFileStorage alloc] init];
        self.serialQueue = dispatch_queue_create("io.naza.analytics.queue.integration", DISPATCH_QUEUE_SERIAL);
        
        [self loadIdentities];
        [self loadMessages];
        [self loadClientId];
        
        self.httpClient = [[NzHttpClient alloc] initWithBaseUrl:self.configuration.endpoint];
        

        [self startFlushTimer];
        
        NSNotificationCenter * nc = [NSNotificationCenter defaultCenter];
        for (NSString *name in @[ UIApplicationDidEnterBackgroundNotification,
                                  UIApplicationDidFinishLaunchingNotification,
                                  UIApplicationWillEnterForegroundNotification,
                                  UIApplicationWillTerminateNotification,
                                  UIApplicationWillResignActiveNotification,
                                  UIApplicationDidBecomeActiveNotification ]) {
            [nc addObserver:self selector:@selector(handleAppStateNotification:) name:name object:nil];
        }
    }
    return self;
}

- (void)setIdentities:(NSArray *) identities {
    Boolean isOne = NO;
    if ([self.cacheIdentities count] == 1 && [kIdentityTypeAnonymousId isEqualToString:((NzIdentity *)[self.cacheIdentities objectAtIndex:0]).type] ) {
        isOne = YES;
    } else {
        for (int i = 0; i < identities.count; i++) {
            if ([self containsIdentity: [identities objectAtIndex:i]] == YES) {
                isOne = YES;
                break;
            }
        }
    }
    
    if (isOne == YES) {
        [self addIdentities:identities];
    } else {
        self.cacheIdentities = [NSMutableArray arrayWithCapacity:identities.count];
        self.messagesQueue = [NSMutableArray arrayWithCapacity:100];
        [self persistIdenitites];
        [self persistMessages];
        [self addIdentities:identities];
        
    }
}

- (void)addIdentities:(NSArray *) identities {
    NSMutableArray * newContactIdentities = [NSMutableArray array];
    for (int i = 0; i < identities.count; i++) {
        NzIdentity * newId = [identities objectAtIndex:i];
        if ([self containsIdentity:newId] == NO) {
            [self.cacheIdentities addObject:newId];
            if ([newId.type isEqualToString:self.configuration.channelTypeId]) {
                [newContactIdentities addObject:newId];
            }
        }
    }
    [self persistIdenitites];
    NSMutableArray * messages = [NSMutableArray array];
    if (newContactIdentities.count == 0) {
        [messages addObject:[NSDictionary dictionary]];
    } else {
        [newContactIdentities enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            [messages addObject:[NzUtils buildMessageWithDomain:kMessageDomainIdentity action:kMessageDomainIdentityActionIdentify data:[obj toDictionary]]];
        }];
    }
    [self sendMessages:messages];
}

- (void)cleanIdentities {
    self.cacheIdentities = [NSMutableArray arrayWithCapacity:10];
    [self persistIdenitites];
}

- (void)sendMessages:(NSArray *)messages {
    [self.messagesQueue addObjectsFromArray:messages];
    if ([self.messagesQueue count] > self.configuration.maxQueueSize) {
        [self.messagesQueue removeObjectsInRange:NSMakeRange(0, 100)];
    }
    [self flushMessages:NO];
}

- (NSString *)generateUUID
{
    return [NzUtils UUID];
}

- (void) startFlushTimer
{
    if (self.flushTimer == nil) {
        self.flushTimer = [NSTimer scheduledTimerWithTimeInterval:self.configuration.flushInterval
                                                           target:self
                                                         selector:@selector(flush)
                                                         userInfo:nil
                                                          repeats:YES];
    }
}

- (void) stopFlushTimer
{
    if (self.flushTimer) {
        [self.flushTimer invalidate];
        self.flushTimer = nil;
    }
}

- (BOOL) verifyNetworkPolicy
{
    //    NotReachable = 0,
    //    ReachableViaWiFi,
    //    ReachableViaWWAN
    if ([[NzAnalyticsNetwork shared] getNetworkStatus] == NzReachableViaWiFi) {
        return YES;
    } else if ([[NzAnalyticsNetwork shared] getNetworkStatus] == NzReachableViaWWAN &&
               self.configuration.networkPolicy == kNetworkPolicyReachable) {
        return YES;
    } else {
        return NO;
    }
}

- (void) flush
{
    [self flushMessages:YES];
}


- (void) flushMessages:(BOOL)force
{
    nz_dispatch_async(self.serialQueue, ^{
       
        if ([self verifyNetworkPolicy] == NO) {
            // 网络策略校验失败
            NzLog(@"flush messages serial queue stopped - network policy is not fulfillment");
            return;
        }
        
        if (self.requestTask != nil) {
            // 有正在发送的请求
            NzLog(@"flush messages serial queue stopped - one request task is running");
            return;
        }
        
        if ([self.messagesQueue count] == 0) {
            // 没有要发送的消息
            NzLog(@"flush messages serial queue stopped - no message queue to send");
            return;
        }
        
        if(force == NO && [self.messagesQueue count] < self.configuration.flushBulkSize) {
            // 非强制发送，并且队列消息数量未达到批量数量阈值
            NzLog(@"flush messages serial queue stopped - no message count is not reached batch size");
            return;
        }
        
        NzLog(@"flush messages serial queue processing");
        
        if([self.cacheIdentities count] == 0) {
            [self.cacheIdentities addObject:[self generateAnonymousId]];
            [self persistIdenitites];
        }
        
        NSArray * sendMessages;
        if ([self.messagesQueue count] >= self.configuration.flushBulkSize) {
            sendMessages = [self.messagesQueue subarrayWithRange:NSMakeRange(0, self.configuration.flushBulkSize - 1)];
        } else {
            sendMessages = [NSArray arrayWithArray:self.messagesQueue];
        }
        NSMutableDictionary * payload = [NSMutableDictionary dictionaryWithObjects:[NSArray arrayWithObjects:
                                                                                    self.clientId,
                                                                                    [self generateUUID],
                                                                                    [NzUtils now],
                                                                                    self.configuration.writeKey,
                                                                                    kSdkType,
                                                                                    kSdkVersion,
                                                                                    [NzModel arrayModelToDictionary:self.cacheIdentities],
                                                                                    sendMessages,
                                                                                    nil]
                                                                           forKeys:[NSArray arrayWithObjects:
                                                                                    @"client_id",
                                                                                    @"request_id",
                                                                                    @"send_ts",
                                                                                    @"write_key",
                                                                                    @"sdk_type",
                                                                                    @"sdk_version",
                                                                                    @"identities",
                                                                                    @"messages",
                                                                                    nil]];
        
        dispatch_semaphore_t sema = dispatch_semaphore_create(0);
        self.requestTask = [self.httpClient post: payload
                                            path: kPathMessages
                                          params: [self commonQueryParams]
                               completionHandler:^(NSData * _Nonnull data, NSError * _Nonnull error) {
            if (error) {
                NzLog(@"flush messages serial queue failed, error code [%d]", error.code);
                if (error.code == kErrorDomainIntegrationBizCode) {
                    // 业务数据问题，抛弃数据
                    [self.messagesQueue removeObjectsInArray:sendMessages];
                    [self persistMessages];
                }
            } else {
                NzLog(@"flush messages serial queue success");
                [self.messagesQueue removeObjectsInArray:sendMessages];
                [self persistMessages];
            }
            self.requestTask = nil;
            dispatch_semaphore_signal(sema);
        }];
        dispatch_semaphore_wait(sema, DISPATCH_TIME_FOREVER);
    });
}

- (void)loadIdentities
{
    NSArray * storedIdentites = [self.storage arrayForKey:kFileStorageIdentities];
    self.cacheIdentities = [NSMutableArray array];
    if (storedIdentites) {
        for (int i = 0; i < [storedIdentites count]; i++) {
            [self.cacheIdentities addObject:[[NzIdentity alloc] initWithDictionary:[storedIdentites objectAtIndex:i]]];
        }
    }
}

- (void)persistIdenitites
{
    [self.storage setArray:[NzModel arrayModelToDictionary:self.cacheIdentities] forKey:kFileStorageIdentities];
}

- (void)clearIdentities
{
    [self.cacheIdentities removeAllObjects];
    [self persistIdenitites];
}

- (NzIdentity *)generateAnonymousId
{
    return [[NzIdentity alloc] initWithId:[self generateUUID] type:kIdentityTypeAnonymousId];
}

- (void)loadMessages
{
    NSArray * storedMessages = [self.storage arrayForKey:kFileStorageMessages];
    self.messagesQueue = [NSMutableArray arrayWithCapacity:self.configuration.maxQueueSize + 1];
    if(storedMessages) {
        [self.messagesQueue addObjectsFromArray:storedMessages];
    }
}

- (void)persistMessages
{
    [self.storage setArray:self.messagesQueue forKey:kFileStorageMessages];
}

- (void)clearMessages
{
    [self.messagesQueue removeAllObjects];
    [self persistMessages];
}

- (NSMutableDictionary *)commonQueryParams
{
    return [NSMutableDictionary dictionaryWithObjects:@[self.configuration.writeKey, self.clientId] forKeys:@[@"token", @"client_id"]];
}

- (Boolean) containsIdentity: (NzIdentity*)identity {
    for (int i = 0; i < self.cacheIdentities.count; i++) {
        NzIdentity * curId = (NzIdentity *)[self.cacheIdentities objectAtIndex:i];
        if ([curId.type isEqualToString:identity.type] && [curId.id isEqualToString:identity.id]) {
            return YES;
        }
    }
    return NO;
}

- (void)handleAppStateNotification: (NSNotification *)notificaiton
{
    NSLog(@"NzIntegration >>> handleAppStateNotification name %@", notificaiton.name);
    if ([notificaiton.name isEqualToString:UIApplicationDidFinishLaunchingNotification] ||
        [notificaiton.name isEqualToString:UIApplicationWillEnterForegroundNotification]) {
        [self startFlushTimer];
    } else if ([notificaiton.name isEqualToString:UIApplicationWillTerminateNotification] ||
               [notificaiton.name isEqualToString:UIApplicationDidEnterBackgroundNotification]) {
        [self stopFlushTimer];
        if ([notificaiton.name isEqualToString:UIApplicationDidEnterBackgroundNotification]) {
            [self beginEventFlushTask];
        }
    }
}

-(void)beginEventFlushTask
{
    self.backgroundFlushTask = [[UIApplication sharedApplication] beginBackgroundTaskWithExpirationHandler:^{
        NzLog(@"end enter background flush task");
        [[UIApplication sharedApplication] endBackgroundTask:self.backgroundFlushTask];
        self.backgroundFlushTask = UIBackgroundTaskInvalid;
    }];
    
    NzLog(@"begin enter background flush task");
    [self flush];
}

- (void)loadClientId
{
    NSDictionary * dict = [[NzFileStorage sharedInstance] dictionaryForKey:kFileStorageIntegrationContext];
    if (dict) {
        self.clientId = [dict objectForKey:@"client_id"];
    }
    if (self.clientId == nil) {
        self.clientId = [self generateUUID];
        [[NzFileStorage sharedInstance] setDictionary:@{@"client_id": self.clientId} forKey:kFileStorageIntegrationContext];
    }
}

- (void)dealloc
{
    [self stopFlushTimer];
}
@end
