//
//  NzFileStorage.m
//  NzAnalytics
//
//  Created by FY on 2018/10/12.
//  Copyright © 2018 FY. All rights reserved.
//

#import "NzFileStorage.h"


static NzFileStorage * sharedInstance;

@interface NzFileStorage ()

@property (nonatomic, strong, nonnull) NSURL *folderURL;

+ (NSURL *_Nullable)applicationSupportDirectoryURL;
+ (NSURL *_Nullable)cachesDirectoryURL;

- (NSURL *_Nonnull)urlForKey:(NSString *_Nonnull)key;
- (void)setData:(NSData *_Nullable)data forKey:(NSString *_Nonnull)key;
- (NSData *_Nullable)dataForKey:(NSString *_Nonnull)key;

@end

@implementation NzFileStorage

+ (instancetype) sharedInstance
{
    if (sharedInstance == nil) {
        sharedInstance = [[NzFileStorage alloc] init];
    }
    return sharedInstance;
}

- (instancetype)init
{
    return [self initWithFolder:[NzFileStorage applicationSupportDirectoryURL]];
}

- (instancetype)initWithFolder:(NSURL *)folderURL
{
    if (self = [super init]) {
        _folderURL = folderURL;
        [self createDirectoryAtURLIfNeeded:folderURL];
        return self;
    }
    return nil;
}

- (void)removeKey:(NSString *)key
{
    NSURL *url = [self urlForKey:key];
    NSError *error = nil;
    [[NSFileManager defaultManager] removeItemAtURL:url error:&error];
}

- (void)resetAll
{
    NSError *error = nil;
    [[NSFileManager defaultManager] removeItemAtURL:self.folderURL error:&error];
    [self createDirectoryAtURLIfNeeded:self.folderURL];
}

- (void)setData:(NSData *)data forKey:(NSString *)key
{
    NSURL *url = [self urlForKey:key];
    
    // a nil value was supplied, remove the storage for said key.
    if (data == nil) {
        [[NSFileManager defaultManager] removeItemAtURL:url error:nil];
        return;
    }
    
    [data writeToURL:url atomically:YES];
    
    
    NSError *error = nil;
    [url setResourceValue:@YES
                   forKey:NSURLIsExcludedFromBackupKey
                    error:&error];
    
}

- (NSData *)dataForKey:(NSString *)key
{
    NSURL *url = [self urlForKey:key];
    NSData *data = [NSData dataWithContentsOfURL:url];
    if (!data) {
        return nil;
    }
    return data;
}

- (nullable NSDictionary *)dictionaryForKey:(NSString *)key
{
    return [self jsonForKey:key];
}

- (void)setDictionary:(nullable NSDictionary *)dictionary forKey:(NSString *)key
{
    [self setJSON:dictionary forKey:key];
}

- (nullable NSArray *)arrayForKey:(NSString *)key
{
    return [self jsonForKey:key];
}

- (void)setArray:(nullable NSArray *)array forKey:(NSString *)key
{
    [self setJSON:array forKey:key];
}

- (nullable NSString *)stringForKey:(NSString *)key
{
    NSDictionary *data = [self jsonForKey:key];
    if (data) {
        return data[key];
    }
    return nil;
}

- (void)setString:(nullable NSString *)string forKey:(NSString *)key
{
    [self setJSON:string forKey:key];
}


+ (NSURL *)applicationSupportDirectoryURL
{
    NSArray *paths = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory, NSUserDomainMask, YES);
    NSString *storagePath = [paths firstObject];
    return [NSURL fileURLWithPath:storagePath];
}

+ (NSURL *)cachesDirectoryURL
{
    NSArray *paths = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
    NSString *storagePath = [paths firstObject];
    return [NSURL fileURLWithPath:storagePath];
}

- (NSURL *)urlForKey:(NSString *)key
{
    return [self.folderURL URLByAppendingPathComponent:key];
}

#pragma mark - Helpers

- (id _Nullable)jsonForKey:(NSString *)key
{
    id result = nil;
    
    NSData *data = [self dataForKey:key];
    if (data) {
        BOOL needsConversion = NO;
        result = [self jsonFromData:data needsConversion:&needsConversion];
        if (needsConversion) {
            [self setJSON:result forKey:key];
            // maybe a little repetitive, but we want to recreate the same path it would
            // take if it weren't being converted.
            data = [self dataForKey:key];
            result = [self jsonFromData:data needsConversion:&needsConversion];
        }
    }
    return result;
}

- (void)setJSON:(id _Nonnull)json forKey:(NSString *)key
{
    NSDictionary *dict = nil;
    
    // json doesn't allow stand alone values like plist (previous storage format) does so
    // we need to massage it a little.
    if (json) {
        if ([json isKindOfClass:[NSDictionary class]] || [json isKindOfClass:[NSArray class]]) {
            dict = json;
        } else {
            dict = @{key: json};
        }
    }
    
    NSData *data = [self dataFromJSON:dict];
    [self setData:data forKey:key];
}

- (NSData *_Nullable)dataFromJSON:(id)json
{
    if (json == nil) {
        return nil;
    }
    
    NSError *error = nil;
    NSData *data = [NSJSONSerialization dataWithJSONObject:json options:0 error:&error];
    return data;
}

- (id _Nullable)jsonFromData:(NSData *_Nonnull)data needsConversion:(BOOL *)needsConversion
{
    NSError *error = nil;
    id result = [NSJSONSerialization JSONObjectWithData:data options:NSJSONReadingMutableContainers error:&error];
    if (error) {
        // maybe it's a plist and needs to be converted.
        result = [self plistFromData:data];
        if (result != nil) {
            *needsConversion = YES;
        }
    }
    return result;
}

- (void)createDirectoryAtURLIfNeeded:(NSURL *)url
{
    if (![[NSFileManager defaultManager] fileExistsAtPath:url.path
                                              isDirectory:NULL]) {
        NSError *error = nil;
        [[NSFileManager defaultManager] createDirectoryAtPath:url.path
                                  withIntermediateDirectories:YES
                                                   attributes:nil
                                                        error:&error];
    }
}

/// Deprecated
- (NSData *_Nullable)dataFromPlist:(nonnull id)plist
{
    NSError *error = nil;
    NSData *data = nil;
    // Temporary just-in-case fix for issue #846; Follow-on PR to move away from plist storage.
    @try {
        data = [NSPropertyListSerialization dataWithPropertyList:plist
                                                          format:NSPropertyListXMLFormat_v1_0
                                                         options:0
                                                           error:&error];
    } @catch (NSException *e) {
        NSLog(@"Unable to serialize data from plist object; Exception: %@, plist: %@", e, plist);
    }
    return data;
}

/// Deprecated
- (id _Nullable)plistFromData:(NSData *_Nonnull)data
{
    NSError *error = nil;
    id plist = [NSPropertyListSerialization propertyListWithData:data
                                                         options:NSPropertyListMutableContainers
                                                          format:nil
                                                           error:&error];
    return plist;
}
@end
